#
# Netri control object, maintaining various bits of state information
#

package Netri;

use strict;
use File::Spec::Functions;
use File::Basename;
use File::Rsync;
use Config::Tiny;
use Hash::Merge qw(merge);
use Make::Target;
use File::Touch;
use Carp;
use Cwd;

BEGIN { Hash::Merge::set_behavior('RIGHT_PRECEDENT') }

# Define labels for levels
my @LEVEL = qw(global hostgroup hostdir);

# ------------------------------------------------------------------------

# Glob a hostlist from a set of files or directories, based on context
sub glob_hostlist
{
  my $self = shift;
  my ($context) = @_;

  $self->{hostlist} = [];
  if ($context eq 'root') {
    for (glob(catfile($self->dir_global, "*"))) {
      next unless -d $_;
      next if $_ eq 'common';    # backwards compatibility with /Net/common
      push @{$self->{hostlist}}, basename($_);
    }
  }

  elsif ($context eq 'hostgroup') {
    $self->{hostlist} = [ map { my $f = basename $_; $f =~ s/^=//; $f } 
      glob(catfile($self->dir_hostgroup, "=*")) ];
  }

  else {
    croak "bad context '$context'";
  }

  return $self->{hostlist};
}

# Get/set the set of hosts to operate over
sub hostlist
{
  my $self = shift;
  return $self->glob_hostlist('root') if $self->level == 0;
  return $self->{hostlist} if $self->{hostlist};

  # Take our hostlist from args, if given
  my $argv = $self->argv;
  if ($argv && @$argv) {
    if (@$argv == 1 && lc $argv->[0] eq 'all') {
      return $self->glob_hostlist('hostgroup');
    }
    else {
      $self->{hostlist} = $self->argv;
    }
  }

  # If no args, use config DefaultHosts, if set
  elsif (my $dh = $self->config->{DefaultHosts}) {
    if (ref $dh eq 'ARRAYREF') {
      $self->{hostlist} = $dh;
    } elsif (! ref $dh) {
      $self->{hostlist} = [ split /[^-\w\.]+/, $dh ];
    } else {
      die "bad config DefaultHosts entry";
    }
  }

  # If no args and no DefaultHosts set, check if only a single host
  else {
    $self->glob_hostlist('hostgroup');
    # If a single-host, default to that
    if (@{$self->{hostlist}} == 1) {
      return $self->{hostlist};
    } 
    # If not a single-host, reset the hostlist
    else {
      $self->{hostlist} = [];
    }
  }

  # Make sure we end up with an arrayref in hostlist
  $self->{hostlist} = [] 
    unless $self->{hostlist} && ref $self->{hostlist} eq 'ARRAY';

  return $self->{hostlist};
}

# Check service restarts, implemented by xsync service-restart.d targets
sub check_service_restarts
{
  my $self = shift;
  my @restart = split /\s*[,\s]\s*/, $self->config->{ServiceRestart} || 
                                     $self->config->{ServicesRestart} || '';
  return unless @restart;

  my $warn = 0;
  my $updated = 0;
  for my $service (@restart) {
    if (! $self->services->{$service}) {
      print "WARNING: cannot restart '$service' service - no service dependencies defined\n";
      $warn++;
      next;
    }

    # Define and check a Make::Target for the given service
    my $target = Make::Target->new(
      targets => catfile(updir, 'etc', 'xsync', 'service-restart.d', "+$service"),
      deps => [ map { s!^/$self->basedir/!!; $_ } split /\s*[,\s]\s*/, $self->services->{$service} ],
      args => [ $self ],
      do => sub {
        my $self = shift;
        my $ntr = shift;
        touch $self->target;
        my $t = $self->target;
        $t =~ s/\+($service)/$1/;
        touch "$t.tmpl.$_" foreach @{$ntr->{hostlist}};
      },
    );
    my @updated = $target->check;
    if (@updated) {
      print "++ Service '$service' restart configured ...\n" 
        if $self->verbose;
      $updated++;
    }
  }

  print "\n" if $warn || $updated;
}

# ------------------------------------------------------------------------
# Push methods

# Call smudge
sub do_smudge ($$;$)
{
  my $self = shift;
  my ($host, $output, $input) = @_;
  die "[do_smudge] missing output parameter" unless $output;
  $output = "-o $output";
  $input = $input ? "-i $input" : '';
  
  # Find smudge variable definition files
  my $exclude = $self->config->{SmudgeExclude} || '-x CVS -x .svn -x {arch}';
  my @tmpl = ( '.tmpl' );
  push @tmpl, map { ".tmpl.$_" } @{$self->config->{Classes}};
  push @tmpl, ".tmpl.$host";
  my $tmpl = join(' ', map { "-t " . $_ } @tmpl);
  my $configlist = join(' ', @{$self->configlist});
  my $cmd = $self->smudge .  " -v $exclude $tmpl $input $output $configlist";
  print "$cmd\n" if $self->verbose >= 2;
  my $out = qx($cmd);
  print $out unless $self->quiet;
}

# Setup common rsync options
sub setup_rsync
{
  my $self = shift;
  my ($host) = @_;

  # Default rsync options
  my %opts = (
    rsh => 'ssh',
    verbose => 1,
    recursive => 1,
    links => 1,
    times => 1,
  );
  my %antonym = (
    links => 'copy-links',
  );
  $antonym{$antonym{$_}} = $_ foreach keys %antonym;
  # Checksum and omit-dir-times only supported by rsync 2.6.6
  $opts{checksum} = 1         unless $self->bootstrap || $self->config->{RsyncOld};
  $opts{'omit-dir-times'} = 1 unless $self->bootstrap || $self->config->{RsyncOld};
  # Allow RsyncArgs to override or supplement defaults
  if ($self->config->{RsyncArgs}) {
    for (split /\s*[\s,]\s*/, $self->config->{RsyncArgs}) {
      if (m/^-[^-]/) {
        warn "Short argument forms not supported with RsyncArgs - '$_' ignored\n";
        next;
      }
      s/^--//;
      my ($arg, $value) = split /=/, $_, 2;
      $value = 1 unless defined $value;
      $opts{$arg} = $value;
      delete $opts{$antonym{$arg}} if $antonym{$arg};
    }
  }
  $opts{'dry-run'} = 1 unless $self->push;

  # Exclusions
  # TODO: this exclude pattern isn't right - there should be patterns
  #   covering all hosts and all classes, not really hostgroup at all
  my @exclude = ( "'*.tmpl." . $self->hostgroup . "*'" );
  $opts{exclude} = \@exclude;
  my @excludefrom = ();
  for ('/etc/netri/EXCLUDE', '../EXCLUDE', './EXCLUDE', "./EXCLUDE.$host") {
    push @excludefrom, $_ if -f $_ && -r $_;
  }
  $opts{'exclude-from'} = \@excludefrom;

  return File::Rsync->new(\%opts);
}

# Do the actual rsync execute with the given options
sub exec_rsync
{
  my $self = shift;
  my ($rsync, $opts) = @_;

  if ($self->verbose) {
    my $cmd = $rsync->getcmd($opts);
    print join(' ', @$cmd) . "\n";
  }
  $rsync->exec($opts) or die "rsync failed";
  my @out = ($rsync->out, $rsync->err);
  @out = grep ! m:^(|building.*done|sent\b.*bytes/sec|total size.*\d+)$:, @out;
  print join '', @out;
}

#rsync $Test $RsyncArgs --exclude='*.tmpl.$Class*' $ExcludeTarg --exclude-from=./EXCLUDE --exclude-from=../EXCLUDE . $PushUser@$targ:$DestDir | egrep -v $RsyncGrepOmit
sub push_hostdir_main
{
  my $self = shift;
  my ($rsync, $host) = @_;

  printf "++ Main sync of %s\n", $self->basedir unless $self->quiet;

  # TODO: basedir isn't right here - it should be destdir, to cope with mappings
  my $opts = { 
    src => '.', 
    dest => sprintf '%s@%s:%s', $self->config->{PushUser}, $host, 
      '/' . $self->basedir,
  };
  $self->exec_rsync($rsync, $opts);

  print "\n" unless $self->quiet;
}

sub push_hostdir_smudged
{
  my $self = shift;
  my ($rsync, $host, $src, $dest) = @_;

  # TODO: this exclude pattern isn't right - there should be patterns
  #   covering all hosts and all classes, not really hostgroup at all
  my @exclude = ( "+ xsync", "*.tmpl." . $self->hostgroup . "*" );
  my $opts = { 
    exclude => \@exclude,
    src => $src,
    dest => sprintf('%s@%s:%s', $self->config->{PushUser}, $host, $dest),
  };
  $self->exec_rsync($rsync, $opts);
}

# rsync $Test $RsyncArgs --delete --include=xsync --exclude='*.tmpl.$Class*' --exclude-from=./EXCLUDE --exclude-from=../EXCLUDE xsync $PushUser@$target:/etc | egrep -v $RsyncGrepOmit
sub push_hostdir_etc
{
  my $self = shift;
  my ($rsync, $host) = @_;
  print "++ Main sync of etc/xsync\n" unless $self->quiet;

  # TODO: this exclude pattern isn't right - there should be patterns
  #   covering all hosts and all classes, not really hostgroup at all
  my @exclude = ( "+ xsync", "*.tmpl." . $self->hostgroup . "*" );
  my $opts = { 
    'delete' => 1,
    exclude => \@exclude,
    src => 'xsync', 
    dest => sprintf('%s@%s:%s', $self->config->{PushUser}, $host,
      '/' . $self->basedir),
  };
  $self->exec_rsync($rsync, $opts);

  print "\n" unless $self->quiet;
}

# Run xsync scripts and restart any services
sub host_scripts
{
  my $self = shift;
  my ($host) = @_;

  return unless $self->xsync;
  return if $self->bootstrap or ! $self->push;

  print "++ Host script execution\n" unless $self->quiet;

  # Do an xsync --scripts
  my $ssh_host = sprintf '%s@%s', $self->config->{SSHUser}, $host;
  my $xsync = $self->xsync;
  $self->exec(qq(ssh $ssh_host $xsync --scripts), 2);

  # Do an xsync --restart, if required
  if (glob("../etc/xsync/service-restart.d/[a-z]*") || 
      glob("../etc/+$host/xsync/service-restart.d/*")) {
    $self->exec(qq(ssh $ssh_host $xsync --restart), 2);
    # Remove per-host services we have restarted
    for my $service (glob("../etc/+$host/xsync/service-restart.d/*")) {
      unlink $service;
      $service =~ s!/\+$host/!/!;
      unlink "$service.tmpl.$host";
    }
  }

  print "\n" unless $self->quiet;
}

# ------------------------------------------------------------------------
# Dispatch methods

sub exec
{
  my $self = shift;
  my ($cmd, $verbosity) = @_;
  $verbosity ||= 0;
  chomp $cmd;
  print qq($cmd\n") if $self->verbose >= $verbosity;
  print qx($cmd);
}

# Pre-push bootstrapping - install BootstrapPackages, setup PushUser
sub bootstrap_pre_push
{
  my $self = shift;
  my ($host) = @_;

  # Install bootstrap packages
  my $test = $self->{noop} ? '--test' : '';
  my $url = $self->config->{BootstrapURL};
  if ($url) {
    print "+ Initial install of bootstrap packages\n";
    for my $pkg (split /\s*[\s,]\s*/, $self->config->{BootstrapPackages}) {
      $self->exec(qq(ssh root\@$host rpm $test -Uvh $url/$pkg));
    }
  }

  # TODO: extra setup magic for non-root PushUsers
}

# Post-push bootstrapping - run a full xsync
sub bootstrap_post_push
{
  my $self = shift;
  my ($host, $rsync) = @_;

  if (my $xsync = $self->xsync) {
    print "+ Bootstrapping packages and services\n";
    my $ssh_user = $self->config->{SSHUser};
    my $test = $self->{noop} ? '-n' : '';
    # Do an xsync --scripts first for bootstrap scripts
    $self->exec(qq(ssh $ssh_user\@$host "$xsync -v $test --scripts"));
    # Do an full xsync to install/startup bootstrap packages and services
    $self->exec(qq(ssh $ssh_user\@$host "$xsync -v $test"));
    # Do another full sync to update any defaults changed e.g. by pkg install
    my @exclude = ( "+ xsync", "*.tmpl." . $self->hostgroup . "*" );
    my $opts = { 
      exclude => \@exclude,
      src => '.', 
      dest => sprintf '%s@%s:%s', $self->config->{PushUser}, $host, 
        '/' . $self->basedir,
    };
    $self->exec_rsync($rsync, $opts);
  }
}

sub push_hostdir
{
  my $self = shift;

  die "No DefaultHosts value set - either set one or specify a host explicitly.\n"
    unless @{$self->hostlist};

  for my $host (@{$self->hostlist}) {
    print "+ [ $host ]\n";

    # Load per-class and per-host configs
    $self->reset_configs;
    # Load host config, specifically for any {Classes} setting
    $self->include(catfile(updir, "=$host"), 'ignore');
    # Load class configs
    if (my $classes = $self->config->{Classes}) {
      $self->config->{Classes} = [ split /\s*[\s,]\s*/, $self->config->{Classes} ]
        unless ref $self->config->{Classes} eq 'ARRAY';
      for (@{$self->config->{Classes}}) {
        $self->include(catfile(updir, updir, "class-$_.ntr"));
      }
    }
    # Reload host config to override any per-host values again
    $self->include(catfile(updir, "=$host"));
    $self->config->{Classes} = [ split /\s*[\s,]\s*/, $self->config->{Classes} ]
      unless ref $self->config->{Classes} eq 'ARRAY';

    my $rsync = $self->setup_rsync($host);

    # Pre-push bootstrapping
    $self->bootstrap_pre_push($host) if $self->bootstrap && $self->level == 2;

    # Do service restart checks
    $self->check_service_restarts();

    # Do main sync
    $self->push_hostdir_main($rsync, $host);

    # Check for a per-hostdir sync
    my $push_hostdir_basedir = "push_hostdir_" . $self->basedir;
    if ($self->can($push_hostdir_basedir)) {
      $self->$push_hostdir_basedir($rsync, $host);
    }

    # Run smudge and do a host-specific smudge sync
    if ($self->smudge) {
      printf "++ Host-specific sync of %s\n", $self->basedir 
        unless $self->quiet;
      $self->do_smudge($host, "+$host");
      $self->push_hostdir_smudged($rsync, $host, "+$host/", '/' . $self->basedir);
      print "\n" unless $self->quiet;

      # For non-etc trees, we also need to smudge the etc service restart dir
      if ($self->basedir ne 'etc' && ! $self->bootstrap) {
        my $restart_dir = catfile(updir, 'etc', 'xsync', 'service-restart.d');
        if (-d $restart_dir) {
          printf "++ Host-specific sync of etc\n" unless $self->quiet;
          my $restart_smudged = catfile(updir, 'etc', "+$host", 'xsync', 'service-restart.d');
          $self->do_smudge($host, $restart_smudged, $restart_dir);
          $self->push_hostdir_smudged($rsync, $host, 
            "$restart_smudged/",
            catfile(rootdir, 'etc', 'xsync', 'service-restart.d'));
          print "\n" unless $self->quiet;
        }
      }
    }

    # Run host-level stuff only if this is a hostdir push
    if ($self->level == 2) {
      $self->host_scripts($host);

      # Post-push bootstrapping
      $self->bootstrap_post_push($host, $rsync) if $self->bootstrap;
    }
  }
}

sub push_hostgroup
{
  my $self = shift;
}

sub push_global
{
  my $self = shift;
}

sub do_push
{
  my $self = shift;
  
  # Do main push
  my $method = "push_" . @LEVEL[$self->level];
  print "+++ dispatching via $method\n\n" if $self->verbose >= 3;
  $self->$method;

  # Post-push bootstrapping
  if ($self->bootstrap) {
  }
}

# ------------------------------------------------------------------------
# Package housekeeping

# Include specified config file, if exists
sub include
{
  my $self = shift;
  my ($conf, $ignore) = @_;
  
  # Glob support
  if ($conf =~ m/\*/) {
    $self->include($_, $ignore) foreach glob($conf);
    return;
  }

  # Standard config
  if (-r $conf) {
    push @{$self->configlist}, $conf unless $ignore;
    my $Config = Config::Tiny->read($conf);
    # Merge standard sections
    $self->config(merge($self->config, $Config->{_})) 
      if keys %{$Config->{_}};
    $self->services(merge($self->services, $Config->{services})) 
      if keys %{$Config->{services}};
  }
}

# Load configs
sub include_configs
{
  my $self = shift;

  # System-wide /etc/netri/netri.conf
  $self->include(catfile(rootdir, "etc", "netri", "netri.conf"));

  # Level 0 (global)
  $self->include(catfile((updir) x $self->level, "global.aap"));
  $self->include(catfile((updir) x $self->level, "global.ntr"));

  # Level 1 (hostgroup)
  $self->include(catfile((updir) x ($self->level-1), "host.aap"));
  $self->include(catfile((updir) x ($self->level-1), "hostgroup.ntr"));

  # Level 2 (hostdir)
  $self->include(catfile((updir) x ($self->level-2), "local.ntr"));

  # Snapshot config trees for resets
  $self->saved_config(merge({}, $self->config));
  $self->saved_services(merge({}, $self->services));
  $self->saved_configlist([ @{$self->configlist} ]);
}

# Reset configs to saved snapshots
sub reset_configs
{
  my $self = shift;
  $self->config(merge({}, $self->saved_config));
  $self->services(merge({}, $self->saved_services));
  $self->configlist([ @{$self->saved_configlist} ]);
}

# Setup methods for top-level attributes
sub make_methods
{
  my $self = shift;

  for (qw(argv basedir bootstrap config configlist 
          dir_global dir_hostdir dir_hostgroup 
          hostgroup level push quiet 
          saved_config saved_configlist saved_services 
          services smudge verbose xsync)) {
    my $code = <<"EOD";
sub $_ {
  my \$self = shift;
  \@_ > 1 and die "too many arguments to $_()";
  \@_ and \$self->{$_} = \$_[0];
  \$self->{$_}
}
EOD

    eval $code;
    die "eval failed: $@" if $@;
  }
}

sub init
{
  my $self = shift;

  # Setup attribute methods
  $self->make_methods;

  # Identify location
  $self->dir_global(undef);
  my @dir = File::Spec->splitdir(cwd);
  my ($dir, $dir1, $dir2);
  my $level = 0;
  do {
    $dir2 = $dir1 if $dir1;
    $dir1 = $dir if $dir;
    $dir = catdir(@dir);
    if (-r catfile($dir, "global.ntr") || -r catfile($dir, "global.aap")) {
      $self->dir_global($dir);
      $self->dir_hostgroup($dir1) if $dir1;
      $self->dir_hostdir($dir2) if $dir2;
      $self->level($level);
    }
    pop @dir if @dir;
    $level++;
  } while ! defined $self->dir_global && @dir;
  die "cannot find 'global.ntr' or 'global.aap' - aborting\n" 
    unless $self->dir_global;
  $self->hostgroup(basename($self->dir_hostgroup)) if $self->dir_hostgroup;
  $self->basedir(basename($self->dir_hostdir)) if $self->dir_hostdir;

  # chdir to hostdir root if level > 2
  if ($self->level > 2) {
    $self->level(2);
    chdir $self->dir_hostdir
      or die "cannot chdir to " . $self->dir_hostdir . ": $!\n";
  }

  # Load user defaults
  my $dot_netri = catfile($ENV{HOME}, '.netri');
  if (-r $dot_netri) {
    $self->{user_config} = Config::Tiny->read($dot_netri);
  }

  # Load location configs
  $self->include_configs;

  # Locate smudge and smudge
  if (my $smudge = qx(which smudge)) {
    chomp $smudge;
    $self->smudge($smudge);
  }
  # TODO: need a way to explicitly disable xsync
  if (my $xsync = $self->config->{Xsync} || qx(which xsync)) {
    chomp $xsync;
    $self->xsync($xsync);
  }

  return $self;
}

sub new
{
  my $class = shift;
  my $self = bless { 
    @_, 
    # Default config options
    config => { 
      SSHUser => 'root', PushUser => 'root', 
      RsyncRsh => 'ssh', 
    },
    services => {},
    configlist => [],
  }, $class;
  $self->init;
}

# ------------------------------------------------------------------------

1;

# arch-tag: edd975d9-a1f3-4485-b425-57c3b2690b39
